#!/usr/bin/env python3
import os,sys,glob,multiprocessing,time,csv,math,pprint
from parsl.app.app import python_app
from os.path import *
from mappgene.subscripts import *

@python_app(executors=['worker'])
def run_ivar(params):

    subject_dir = params['work_dir']
    subject = basename(subject_dir)
    input_reads = params['input_reads']
    variant_frequency = params['variant_frequency']
    stdout = params['stdout']
    ivar_dir = join(subject_dir, 'ivar')
    output_dir = join(subject_dir, 'ivar_outputs')
    alignments_dir = join(output_dir, 'alignments')
    raw_dir = join(ivar_dir, 'raw_data')
    smart_remove(raw_dir)
    smart_remove(output_dir)
    smart_mkdir(raw_dir)
    smart_mkdir(output_dir)
    smart_mkdir(alignments_dir)
    reads = []

    start_time = time.time()
    start_str = f'''
=====================================
Starting iVar with subject: {subject}
{get_time_date()}
Arguments: 
{pprint.pformat(params, width=1)}
=====================================
'''
    write(stdout, start_str)
    print(start_str)
    update_permissions(ivar_dir, params)
    update_permissions(output_dir, params)

    # Run fixq.sh
    for input_read in input_reads:
        tmp_f = join(raw_dir, 'tmp_' + basename(input_read))
        f = join(raw_dir, basename(input_read))
        smart_copy(input_read, f)
        run(f'zcat {f} | awk \'NR%4 == 0 {{ gsub(\\"F\\", \\"?\\"); gsub(\\":\\", \\"5\\") }}1\'' +
            f' | gzip -c > {tmp_f}', params)
        if exists(tmp_f):
            smart_remove(f)
            os.rename(tmp_f, f)
        reads.append(f)

    # Deinterleave if only a single FASTQ was found
    # fasta = join(ivar_dir, 'references/PS_1200bp.fasta')
    fasta = join(ivar_dir, 'references/NC_045512.2.fasta')
    if len(reads) == 1:
        f = reads[0]
        read1 = replace_extension(f, '_R1.fastq.gz')
        read2 = replace_extension(f, '_R2.fastq.gz')
        run(f'reformat.sh in={f} out1={read1} out2={read2}', params)
        smart_remove(f)
    elif len(reads) == 2:
        reads.sort()
        read1 = reads[0]
        read2 = reads[1]
    else:
        raise Exception(f'Invalid reads: {reads}')

    
    align_prefix = join(alignments_dir, subject)
    bam = replace_extension(align_prefix, '.bam')
    trimmed = replace_extension(align_prefix, '.trimmed')
    trimmed_sorted = replace_extension(align_prefix, '.trimmed.sorted.bam')
    variants = replace_extension(align_prefix, '.variants')
    noinsertions = replace_extension(align_prefix, '.noins.variants')
    masked = replace_extension(align_prefix, '.masked.txt')
    trimmed_masked = replace_extension(align_prefix, '.trimmed.masked.bam')
    final_masked = replace_extension(align_prefix, '.final.masked.variants')
    lofreq_bam = replace_extension(align_prefix, '.lofreq.bam')
    vcf_s0 = replace_extension(align_prefix, '.vcf')
    tsv = replace_extension(align_prefix, '.final.masked.variants.tsv')
    output_tsv = join(output_dir, f'{subject}.ivar.final.masked.variants.tsv')
    run(f'bwa index {fasta}', params)
    run(f'bwa mem -t 8 {fasta} {read1} {read2} | samtools sort -o {bam}', params)
    run(f'ivar trim -b {ivar_dir}/nCoV-2019.scheme.bed -p {trimmed} -i {bam} -e', params)
    run(f'samtools sort {trimmed}.bam -o {trimmed_sorted}', params)

    # call variants with ivar (produces {subject}.variants.tsv)
    run(f'samtools mpileup -aa -A -d 0 -B -Q 0 {trimmed_sorted} | ' +
        f'ivar variants -p {variants} -q 20 -t {variant_frequency} -r {fasta} ' +
        f'-g {ivar_dir}/GCF_009858895.2_ASM985889v3_genomic.gff', params)
    
    # remove low quality insertions because we want to ignore most mismatches
    # to primers that are insertions (produces {subject}.noins.variants.tsv)
    run(f"awk \'! (\\$4 ~ /^\\+/ && \\$10 >= 20) {{ print }}\' < {variants}.tsv > {noinsertions}.tsv", params)
    
    # get primers with mismatches to reference (produces {subject}.masked.txt)
    run(f'ivar getmasked -i {noinsertions}.tsv -b {ivar_dir}/nCoV-2019.bed ' +
        f'-f {ivar_dir}/nCoV-2019.tsv -p {masked}', params)
    
    # remove reads with primer mismatches (produces {subject}.trimmed.masked.bam)
    run(f'ivar removereads -i {trimmed_sorted} -p {trimmed_masked} ' +
        f'-t {masked} -b {ivar_dir}/nCoV-2019.bed', params)
    
    # call variants with reads with primer mismatches removed (produces {subject}.final.masked.variants)
    run(f'samtools mpileup -aa -A -d 0 -B -Q 0 {trimmed_masked} | ' +
        f'ivar variants -p {final_masked} -q 20 -t {variant_frequency} -r {fasta} ' +
        f'-g {ivar_dir}/GCF_009858895.2_ASM985889v3_genomic.gff', params)
    smart_copy(tsv, output_tsv)

    # # use lofreq to convert bam to vcf
    run(f'lofreq indelqual --dindel -f {fasta} -o {lofreq_bam} --verbose {trimmed_masked}', params)
    run(f'samtools index {lofreq_bam}', params)
    run(f'lofreq call --call-indels -f {fasta} -o {vcf_s0} --verbose {lofreq_bam}', params)

    # Run snpEff postprocessing
    vcf_s1 = join(output_dir, f'{subject}.ivar.vcf')
    vcf_s2 = join(output_dir, f'{subject}.ivar.snpEFF.vcf')
    vcf_s3 = join(output_dir, f'{subject}.ivar.snpSIFT.txt')
    run(f'sed "s/MN908947.3/NC_045512.2/g" {vcf_s0} > {vcf_s1}', params)
    run(f'java -Xmx8g -jar /opt/snpEff/snpEff.jar NC_045512.2 {vcf_s1} > {vcf_s2}', params)
    run(f'cat {vcf_s2} | /opt/snpEff/scripts/vcfEffOnePerLine.pl | java -jar /opt/snpEff/SnpSift.jar ' +
        f' extractFields - CHROM POS REF ALT AF DP "ANN[*].IMPACT" "ANN[*].FEATUREID" "ANN[*].EFFECT" ' +
        f' "ANN[*].HGVS_C" "ANN[*].HGVS_P" "ANN[*].CDNA_POS" "ANN[*].AA_POS" "ANN[*].GENE" > {vcf_s3}', params)

    # Clear extra files
    smart_remove('snpEff_genes.txt')
    smart_remove('snpEff_summary.html')

    update_permissions(ivar_dir, params)
    update_permissions(output_dir, params)
    finish_str = f'''
=====================================
Finished iVar with subject: {subject}
{get_time_date()}
Arguments: 
{pprint.pformat(params, width=1)}
Total time: {get_time_string(time.time() - start_time)} (HH:MM:SS)
=====================================
'''
    write(stdout, finish_str)
    print(finish_str)



#!/usr/bin/env python3
import argparse,parsl,os,sys,glob,shutil
from os.path import *
from mappgene.subscripts import *

script_dir = abspath(os.path.dirname(os.path.realpath(__file__)))
cwd = abspath(os.getcwd())

def parse_args(args):


    parser = argparse.ArgumentParser()

    if not '--test' in sys.argv:
        parser.add_argument('inputs', nargs='+',
            help='Paths to FASTQ input file(s).')

    parser.add_argument('--test', action='store_true',
        help='Test using the example inputs.')

    workflow_group = parser.add_mutually_exclusive_group(required=True)

    workflow_group.add_argument('--ivar', action='store_true',
        help='Run the iVar workflow.')

    workflow_group.add_argument('--vpipe', action='store_true',
        help='Run the V-pipe workflow.')

    parser.add_argument('--outputs', '-o', default='mappgene_outputs/',
        help='Path to output directory.')

    parser.add_argument('--container', default=join(cwd, 'image.sif'),
        help='Path to Singularity container image.')

    parser.add_argument('--read-length', default=130,
        help='V-pipe: read length in sample.tsv (see cbg-ethz.github.io/V-pipe/tutorial/sars-cov2).')

    parser.add_argument('--variant_frequency', default=0.01,
        help='iVar: variant frequency cutoff.')

    scheduler_group = parser.add_mutually_exclusive_group()

    scheduler_group.add_argument('--slurm', action='store_true',
        help='Use the Slurm scheduler.')

    scheduler_group.add_argument('--flux', action='store_true',
        help='Use the Flux scheduler.')

    parser.add_argument('--nnodes', '-n', default=1,
        help='Slurm/Flux: number of nodes.')

    parser.add_argument('--bank', '-b', default='asccasc',
        help='Slurm/Flux: bank to charge for jobs.')

    parser.add_argument('--partition', '-p', default='pbatch',
        help='Slurm/Flux: partition to assign jobs.')

    parser.add_argument('--walltime', '-t', default='11:59:00',
        help='Slurm/Flux: walltime in format HH:MM:SS.')

    return parser.parse_args()

def main():

    args = parse_args(sys.argv[1:])

    # Copy V-pipe repo as main working directory
    tmp_dir = join(cwd, 'tmp')
    vpipe_dir = join(tmp_dir, 'vpipe')
    base_params = {
        'container': abspath(args.container),
        'work_dir': tmp_dir,
        'read_length': args.read_length,
        'variant_frequency': args.variant_frequency,
        'stdout': abspath(join(args.outputs, 'mappgene.stdout')),
    }

    if shutil.which('singularity') is None:
        raise Exception(f"Missing Singularity executable in PATH.\n\n" +
            f"Please ensure Singularity is installed: https://sylabs.io/guides/3.0/user-guide/installation.html")

    if not exists(base_params['container']):
        raise Exception(f"Missing container image at {base_params['container']}\n\n" +
            f"Either specify another image with --container\n\n" +
            f"Or build the container with the recipe at: {join(script_dir, 'data/container/recipe.def')}\n\n" +
            f"Or download the container at: https://www.dropbox.com/s/ymsfn9z7v3utqe0/image.sif?dl=1\n")

    smart_remove(tmp_dir)
    smart_mkdir(tmp_dir)
    
    run(f'cp -rf /opt/vpipe {vpipe_dir}', base_params)
    smart_copy(join(script_dir, 'data/extra_files'), tmp_dir)

    run(f'cd {vpipe_dir} && sh init_project.sh || true', base_params)
    update_permissions(tmp_dir, base_params)

    if args.test:
        args.inputs = join(script_dir, 'data/example_inputs/*.fastq.gz')
    
    if isinstance(args.inputs, str):
        args.inputs = glob(args.inputs)

    all_params = {}

    # Copy reads to subject directory
    for input_read in args.inputs:
        pair1 = input_read.replace('_R2.', '_R1.')
        pair2 = input_read.replace('_R1.', '_R2.')
        if '_R1.' in input_read and pair2 not in args.inputs:
            raise Exception(f'Missing paired read: {pair2}')
        if '_R2.' in input_read and pair1 not in args.inputs:
            raise Exception(f'Missing paired read: {pair1}')

        subject = replace_extension(basename(input_read).replace('_R1.', '.').replace('_R2.', '.'))
        subject_dir = abspath(join(args.outputs, subject))

        if not subject in all_params:
            smart_copy(tmp_dir, subject_dir)
            params = base_params.copy()
            params['work_dir'] = subject_dir
            params['input_reads'] = [input_read]
            params['stdout'] = join(subject_dir, 'worker.stdout')
            all_params[subject] = params
        else:
            all_params[subject]['input_reads'].append(input_read)

    if args.slurm:
        executor = parsl.executors.HighThroughputExecutor(
            label="worker",
            address=parsl.addresses.address_by_hostname(),
            provider=parsl.providers.SlurmProvider(
                args.partition,
                launcher=parsl.launchers.SrunLauncher(),
                nodes_per_block=int(args.nnodes),
                init_blocks=1,
                max_blocks=1,
                worker_init=f"export PYTHONPATH=$PYTHONPATH:{os.getcwd()}",
                walltime=args.walltime,
                scheduler_options="#SBATCH --exclusive\n#SBATCH -A {}\n".format(args.bank),
                move_files=False,
            ),
        )
    elif args.flux:
        executor = parsl.executors.FluxExecutor(
            label="worker",
            flux_path="/usr/global/tools/flux/toss_3_x86_64_ib/flux-c0.28.0.pre-s0.17.0.pre/bin/flux",
            provider=parsl.providers.SlurmProvider(
                args.partition,
                launcher=parsl.launchers.SrunLauncher(),
                nodes_per_block=int(args.nnodes),
                init_blocks=1,
                max_blocks=1,
                worker_init=f"export PYTHONPATH=$PYTHONPATH:{os.getcwd()}",
                walltime=args.walltime,
                scheduler_options="#SBATCH --exclusive\n#SBATCH -A {}\n".format(args.bank),
                move_files=False,
            ),
        )
    else:
        executor = parsl.executors.ThreadPoolExecutor(label="worker")

    config = parsl.config.Config(executors=[executor])
    parsl.set_stream_logger()
    parsl.load(config)

    if args.ivar:
        results =  []
        for params in all_params.values():
            results.append(run_ivar(params))
        for r in results:
            r.result()

    elif args.vpipe:
        results =  []
        for params in all_params.values():
            results.append(run_vpipe(params))
        for r in results:
            r.result()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import os,sys,glob,multiprocessing,time,csv,math,pprint
from parsl.app.app import python_app
from os.path import *
from mappgene.subscripts import *

@python_app(executors=['worker'])
def run_vpipe(params):

    subject_dir = params['work_dir']
    subject = basename(subject_dir)
    input_reads = params['input_reads']
    stdout = params['stdout']
    vpipe_dir = join(subject_dir, 'vpipe')
    output_dir = join(subject_dir, 'vpipe_outputs')
    raw_dir = join(vpipe_dir, 'samples/a/b/raw_data')
    smart_remove(raw_dir)
    smart_remove(output_dir)
    smart_mkdir(raw_dir)
    smart_mkdir(output_dir)
    reads = []

    start_time = time.time()
    start_str = f'''
=====================================
Starting V-pipe with subject: {subject}
{get_time_date()}
Arguments: 
{pprint.pformat(params, width=1)}
=====================================
'''
    write(stdout, start_str)
    print(start_str)
    update_permissions(vpipe_dir, params)
    update_permissions(output_dir, params)

    # Run fixq.sh
    for input_read in input_reads:
        tmp_f = join(raw_dir, 'tmp_' + basename(input_read))
        f = join(raw_dir, basename(input_read))
        smart_copy(input_read, f)
        run(f'zcat {f} | awk \'NR%4 == 0 {{ gsub(\\"F\\", \\"?\\"); gsub(\\":\\", \\"5\\") }}1\'' +
            f' | gzip -c > {tmp_f}', params)
        if exists(tmp_f):
            smart_remove(f)
            os.rename(tmp_f, f)
        reads.append(f)

    # Deinterleave if only a single FASTQ was found
    # fasta = join(vpipe_dir, 'references/PS_1200bp.fasta')
    fasta = join(vpipe_dir, 'references/NC_045512.2.fasta')
    if len(reads) == 1:
        f = reads[0]
        out1 = replace_extension(f, '_R1.fastq.gz')
        out2 = replace_extension(f, '_R2.fastq.gz')
        stat = replace_extension(f, '.stat')
        run(f"bbduk.sh in={f} out1={out1} out2={out2} ref={fasta} stats={stat} " +
            "k=13 ktrim=l hdist=2 restrictleft=31 statscolumns=5 minlen=65", params)
        smart_remove(f)

    # Update sample.tsv with read length
    smart_remove(join(vpipe_dir, '.snakemake'))
    run(f'cd {vpipe_dir} && ./vpipe --dryrun', params)
    samples_tsv = join(vpipe_dir, 'samples.tsv')
    with open(samples_tsv, 'r') as fr, open(f'{samples_tsv}.tmp', 'w') as fw:
        r = csv.reader(fr, delimiter='\t')
        w = csv.writer(fw, delimiter='\t')
        for row in r:
            row.append(params['read_length'])
            w.writerow(row)
    smart_copy(f'{samples_tsv}.tmp', samples_tsv)

    # Run V-pipe analysis
    ncores = int(math.floor(multiprocessing.cpu_count() / 2))
    run(f'cd {vpipe_dir} && ./vpipe --cores {ncores} --use-conda --unlock', params)
    run(f'cd {vpipe_dir} && ./vpipe --cores {ncores} --use-conda', params)
    time.sleep(10)

    # Run snpEff postprocessing
    vcf_s0 = join(vpipe_dir, 'samples/a/b/variants/SNVs/snvs.vcf')
    vcf_s1 = join(output_dir, f'{subject}.vpipe.vcf')
    vcf_s2 = join(output_dir, f'{subject}.vpipe.snpEFF.vcf')
    vcf_s3 = join(output_dir, f'{subject}.vpipe.snpSIFT.txt')
    run(f'sed "s/MN908947.3/NC_045512.2/g" {vcf_s0} > {vcf_s1}', params)
    run(f'java -Xmx8g -jar /opt/snpEff/snpEff.jar NC_045512.2 {vcf_s1} > {vcf_s2}', params)
    run(f'cat {vcf_s2} | /opt/snpEff/scripts/vcfEffOnePerLine.pl | java -jar /opt/snpEff/SnpSift.jar ' +
        f' extractFields - CHROM POS REF ALT AF DP "ANN[*].IMPACT" "ANN[*].FEATUREID" "ANN[*].EFFECT" ' +
        f' "ANN[*].HGVS_C" "ANN[*].HGVS_P" "ANN[*].CDNA_POS" "ANN[*].AA_POS" "ANN[*].GENE" > {vcf_s3}', params)

    # Copy outputs to accessible location
    smart_copy(join(vpipe_dir, 'samples/a/b/alignments'), join(output_dir, 'alignments'))
    smart_copy(join(vpipe_dir, 'samples/a/b/variants'), join(output_dir, 'variants'))
    smart_copy(join(vpipe_dir, 'samples/a/b/extracted_data'), join(output_dir, 'extracted_data'), exclude=['*.gz', '*.fasta'])
    smart_copy(join(vpipe_dir, 'samples/a/b/preprocessed_data'), join(output_dir, 'preprocessed_data'), exclude=['*.gz', '*.fasta'])
    smart_copy(join(vpipe_dir, 'samples/a/b/raw_data'), join(output_dir, 'raw_data'), exclude=['*.gz', '*.fasta'])
    smart_copy(join(vpipe_dir, 'samples/a/b/references'), join(output_dir, 'references'), exclude=['*.gz', '*.fasta'])

    # Clear extra files
    smart_remove('snpEff_genes.txt')
    smart_remove('snpEff_summary.html')

    update_permissions(vpipe_dir, params)
    update_permissions(output_dir, params)
    finish_str = f'''
=====================================
Finished V-pipe with subject: {subject}
{get_time_date()}
Arguments: 
{pprint.pformat(params, width=1)}
Total time: {get_time_string(time.time() - start_time)} (HH:MM:SS)
=====================================
'''
    write(stdout, finish_str)
    print(finish_str)




import sys
from os.path import abspath, split, join
from setuptools import setup, find_packages

src_path = split(abspath(sys.argv[0]))[0]
src_path = join(src_path, 'mappgene')

# ------------------------------------------------------------------------------
setup(
    name='mappgene',
    description='Genomic sequence analysis for high-performance computing',
    version='1.0.0',
    python_requires='>=3.7',
    author='Joseph Moon',
    author_email='moon15@llnl.gov',
    entry_points={
        'console_scripts': [
            'mappgene = mappgene.cli:main'
        ]
    },
    packages=find_packages(),
    include_package_data=True,
    package_data={'mappgene' : ['data/*'] },
    install_requires=['pip>=21.2.4', 'parsl>=1.1.0', 'pytest>=6.2.4'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: System :: Distributed Computing",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)

# ------------------------------------------------------------------------------
